/*
 * Copyright (C) 2008 Michael Brown <mbrown@fensystems.co.uk>.
 * Copyright (C) 2013 Tal Aloni <tal.aloni.il@gmail.com>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ntddk.h>
#include <initguid.h>
#include <wdmsec.h>
#include <ntdddisk.h>
#include <coguid.h>
#include "wait4ufd.h"

/** Maximum time to wait for system disk, in seconds */
#define WAIT4UFD_MAX_WAIT 10

/**
 * Fetch disk signature
 *
 * @v name		Disk device name
 * @v device		Disk device object
 * @v file		Disk file object
 * @v info		Partition information buffer
 * @ret status		NT status
 */
static NTSTATUS fetch_partition_info ( PUNICODE_STRING name,
				       PDEVICE_OBJECT device,
				       PFILE_OBJECT file,
				       PDISK_PARTITION_INFO info ) {
	KEVENT event;
	struct {
		DISK_GEOMETRY_EX geometry;
		DISK_PARTITION_INFO __dummy_partition_info;
		DISK_DETECTION_INFO __dummy_detection_info;
	} buf;
	IO_STATUS_BLOCK io_status;
	PIRP irp;
	PIO_STACK_LOCATION io_stack;
	PDISK_PARTITION_INFO partition_info;
	NTSTATUS status;

	/* Construct IRP to fetch drive geometry */
	KeInitializeEvent ( &event, NotificationEvent, FALSE );
	irp = IoBuildDeviceIoControlRequest ( IOCTL_DISK_GET_DRIVE_GEOMETRY_EX,
					      device, NULL, 0, &buf,
					      sizeof ( buf ), FALSE, &event,
					      &io_status );
	if ( ! irp ) {
		DbgPrint ( "Could not build IRP to retrieve geometry for "
			   "\"%wZ\"\n", name );
		return STATUS_UNSUCCESSFUL;
	}
	io_stack = IoGetNextIrpStackLocation ( irp );
	io_stack->FileObject = file;

	/* Issue IRP */
	status = IoCallDriver ( device, irp );
	if ( status == STATUS_PENDING ) {
		status = KeWaitForSingleObject ( &event, Executive, KernelMode,
						 FALSE, NULL );
	}
	if ( NT_SUCCESS ( status ) )
		status = io_status.Status;
	if ( ! NT_SUCCESS ( status ) ) {
		DbgPrint ( "IRP failed to retrieve geometry for \"%wZ\": %x\n",
			   name, status );
		return status;
	}

	/* Extract partition information */
	partition_info = DiskGeometryGetPartition ( &buf.geometry );
	memcpy ( info, partition_info, sizeof ( *info ) );

	return STATUS_SUCCESS;
}

/**
 * Check for system disk
 *
 * @v name		Disk device name
 * @v boot_info		Boot disk information
 * @ret status		NT status
 */
static NTSTATUS check_system_disk ( PUNICODE_STRING name,
				    PBOOTDISK_INFORMATION_EX boot_info ) {
	BOOLEAN must_disable;
	PFILE_OBJECT file;
	PDEVICE_OBJECT device;
	DISK_PARTITION_INFO info;
	NTSTATUS status;

	/* Enable interface if not already done */
	status = IoSetDeviceInterfaceState ( name, TRUE );
	/* If interface is already enabled, IoSetDeviceInterfaceState
	 * will return STATUS_OBJECT_NAME_EXISTS, which counts as a
	 * success status.
	 */
	must_disable = ( ( NT_SUCCESS ( status ) &&
			   ( status != STATUS_OBJECT_NAME_EXISTS ) )
			 ? TRUE : FALSE );

	/* Get device and file object pointers */
	status = IoGetDeviceObjectPointer ( name, FILE_ALL_ACCESS, &file,
					    &device );
	if ( ! NT_SUCCESS ( status ) ) {
		/* Most probably not yet attached */
		DbgPrint ( "  Disk unavailable (%lx): \"%wZ\"\n",
			   status, name );
		goto err_iogetdeviceobjectpointer;
	}

	/* Get disk signature */
	status = fetch_partition_info ( name, device, file, &info );
	if ( ! NT_SUCCESS ( status ) )
		goto err_fetch_partition_info;

	/* Check for a matching disk signature */
	status = STATUS_UNSUCCESSFUL;
	switch ( info.PartitionStyle ) {
	case PARTITION_STYLE_MBR:
		DbgPrint ( "  MBR %08lx: \"%wZ\"\n",
			   info.Mbr.Signature, name );
		if ( boot_info->SystemDeviceIsGpt )
			goto err_not_system_disk;
		if ( boot_info->SystemDeviceSignature != info.Mbr.Signature )
			goto err_not_system_disk;
		break;
	case PARTITION_STYLE_GPT:
		DbgPrint ( "  GPT " GUID_FMT ": \"%wZ\"\n",
			   GUID_ARGS ( info.Gpt.DiskId ), name );
		if ( ! boot_info->SystemDeviceIsGpt )
			goto err_not_system_disk;
		if ( ! IsEqualGUID ( &boot_info->SystemDeviceGuid,
				       &info.Gpt.DiskId ) )
			goto err_not_system_disk;
		break;
	default:
		DbgPrint ( "  Unhandled disk style %d: \"%wZ\"\n",
			   info.PartitionStyle, name );
		status = STATUS_NOT_SUPPORTED;
		goto err_unknown_type;
	}

	/* Success */
	DbgPrint ( "Found system disk at \"%wZ\"\n", name );
	status = STATUS_SUCCESS;

 err_not_system_disk:
 err_unknown_type:
 err_fetch_partition_info:
	/* Drop object reference */
	ObDereferenceObject ( file );
 err_iogetdeviceobjectpointer:
	/* Disable interface if we had to enable it */
	if ( must_disable )
		IoSetDeviceInterfaceState ( name, FALSE );
	return status;
}

/**
 * Find system disk
 *
 * @ret status		NT status
 */
static NTSTATUS find_system_disk ( VOID ) {
	union {
		BOOTDISK_INFORMATION basic;
		BOOTDISK_INFORMATION_EX extended;
	} boot_info;
	PWSTR symlinks;
	PWSTR symlink;
	UNICODE_STRING u_symlink;
	NTSTATUS status;

	/* Get boot disk information */
	RtlZeroMemory ( &boot_info, sizeof ( boot_info ) );
	status = IoGetBootDiskInformation ( &boot_info.basic,
					    sizeof ( boot_info ) );
	if ( ! NT_SUCCESS ( status ) ) {
		DbgPrint ( "Could not get boot disk information: %x\n",
			   status );
		goto err_getbootdiskinformation;
	}
	if ( boot_info.extended.SystemDeviceIsGpt ) {
		DbgPrint ( "  System disk is GPT " GUID_FMT ",",
			   GUID_ARGS ( boot_info.extended.SystemDeviceGuid ) );
	} else if ( boot_info.extended.SystemDeviceSignature ) {
		DbgPrint ( "  System disk is MBR %08lx,",
			   boot_info.extended.SystemDeviceSignature );
	} else {
		DbgPrint ( "  System disk is <unknown>," );
	}
	if ( boot_info.extended.BootDeviceIsGpt ) {
		DbgPrint ( " boot disk is GPT " GUID_FMT "\n",
			   GUID_ARGS ( boot_info.extended.BootDeviceGuid ) );
	} else if ( boot_info.extended.BootDeviceSignature ) {
		DbgPrint ( " boot disk is MBR %08lx\n",
			   boot_info.extended.BootDeviceSignature );
	} else {
		DbgPrint ( " boot disk is <unknown>\n" );
	}

	/* The only way to access the BootDeviceSignature and SystemDeviceSignature
	 * passed from the boot loader is through IoGetBootDiskInformation().
	 * However, IoGetBootDiskInformation() will enumerate through all of the
	 * disks that are currently present in the system, and will only set 
	 * the appropriate BootDiskInformation variables if the boot / system disk
     * is detected.
	 */
	if ( !boot_info.extended.SystemDeviceIsGpt &&
	      boot_info.extended.SystemDeviceSignature == 0 ) {
		/* System disk signature is not yet available */
		return STATUS_NOT_FOUND;
	}

	/* Enumerate all disks */
	status = IoGetDeviceInterfaces ( &GUID_DEVINTERFACE_DISK, NULL,
					 DEVICE_INTERFACE_INCLUDE_NONACTIVE,
					 &symlinks );
	if ( ! NT_SUCCESS ( status ) ) {
		DbgPrint ( "Could not fetch disk list: %x\n", status );
		goto err_getdeviceinterfaces;
	}

	/* Look for the system disk */
	status = STATUS_NOT_FOUND;
	for ( symlink = symlinks ;
	      RtlInitUnicodeString ( &u_symlink, symlink ) , *symlink ;
	      symlink += ( ( u_symlink.Length / sizeof ( *symlink ) ) + 1 ) ) {
		status = check_system_disk ( &u_symlink, &boot_info.extended );
		if ( NT_SUCCESS ( status ) )
			break;
	}

	/* Free object list */
	ExFreePool ( symlinks );
 err_getdeviceinterfaces:
 err_getbootdiskinformation:
	return status;
}

/**
 * Wait for system disk to appear
 *
 * @v driver		Driver object
 * @v context		Context
 * @v count		Number of times this routine has been called
 */
static VOID wait_for_system_disk ( PDRIVER_OBJECT driver, PVOID context,
			       ULONG count ) {
	LARGE_INTEGER delay;
	NTSTATUS status;

	DbgPrint ( "Waiting for system disk (attempt %ld)\n", count );

	/* Check for existence of system disk */
	status = find_system_disk();
	if ( NT_SUCCESS ( status ) ) {
		DbgPrint ( "Found system disk; proceeding with boot\n" );
		return;
	}

	/* Give up after too many attempts */
	if ( count >= WAIT4UFD_MAX_WAIT ) {
		DbgPrint ( "Giving up waiting for system disk\n" );
		return;
	}

	/* Sleep for a second, reschedule self */
	delay.QuadPart = -10000000L /* 1 second, relative to current time */;
	KeDelayExecutionThread ( KernelMode, FALSE, &delay );
	IoRegisterBootDriverReinitialization ( driver, wait_for_system_disk,
					       context );
}

/**
 * Driver entry point
 *
 * @v DriverObject	Driver object
 * @v RegistryPath	Driver-specific registry path
 * @ret ntstatus	NT status
 */
NTSTATUS DriverEntry ( IN PDRIVER_OBJECT DriverObject,
		       IN PUNICODE_STRING RegistryPath ) {

	DbgPrint ( "Wait4UFD Driver initializing\n" );

	/* Wait for system disk */
	IoRegisterBootDriverReinitialization ( DriverObject, wait_for_system_disk, NULL );

	DbgPrint ( "Wait4UFD Driver finalizing\n" );
	UNREFERENCED_PARAMETER(DriverObject);
	UNREFERENCED_PARAMETER(RegistryPath);
	return STATUS_SUCCESS;
}
